local VERSION = tonumber(("$Revision$"):match("%d+"))
local DURATION = 0.5

local L = AceLibrary("AceLocale-2.2"):new("PitBull_BarFader")
L:RegisterTranslations("enUS", function() return {
	["BarFader"] = true,
	["Ignore Unit(s)"] = true,
	["Smooth Value Changing"] = true,
} end)
local PitBull = PitBull
local PitBull_BarFader = PitBull:NewModule("BarFader", "AceHook-2.1", "AceEvent-2.0")
local _newFrame, _delFrame = PitBull.newFrame, PitBull.delFrame
local _PitBull_Health = PitBull:GetModule("Health")
local _PitBull_Mana = PitBull:GetModule("Mana")
local _wrap = PitBull.wrap
local _frames = {}
local _framesSmooth = {}
local _pi,_cos = math.pi, math.cos
local _muTable = setmetatable({}, {
	__index = function(name,mu)
		local tmp = (1-_cos(mu*_pi))/2
		name[mu] = tmp
		return tmp
	end
})

-- @desc Cosine interpolation.
-- @arg y1 number Start Value
-- @arg y2 number End Value
-- @arg mu number Step Position (0..1)
-- @cat Util
-- stolen from Dreamchat.  all credit goes to them
local function CosineInterpolate(y1, y2, mu)
	local mu2 = _muTable[mu]
	return y1*(1-mu2)+y2*mu2
end

function PitBull_BarFader:UpdateFaders()
	local now, alpha, stop, bar, cVal, sVal, mu = GetTime(), 1
	for bar,field in pairs(_frames) do
		stop = bar.faderstop
		if bar.realValue then
			alpha = 1 - ((stop - now) / DURATION)
		else
			alpha = ((stop - now) / DURATION)
		end
		if (alpha <= 0 and not bar.realValue) or (alpha >= 1 and bar.realValue) then
			if (bar:GetParent()[field] == bar) then -- make sure we didn't get reassigned somehow
				bar:SetAlpha(1)
				bar:Hide()
				bar.orientation = nil
				bar.stop = nil
				if bar.realValue then
					if field == "healthFade" then
						bar:GetParent().healthBar:SetValue(bar.realValue)
					elseif field == "manaFade" then
						bar:GetParent().manaBar:SetValue(bar.realValue)
					end
				end
				bar.realValue = nil
				bar:GetParent()[field] = _delFrame(bar)
			end
			_frames[bar] = nil
		else
			bar:SetAlpha(alpha)
		end
	end
	for bar,field in pairs(_framesSmooth) do
		stop = bar.faderstop
		sVal = bar.smoothstop
		if bar.realValue then
			alpha = 1 - ((stop - now) / DURATION)
			mu = alpha
		else
			alpha = ((stop - now) / DURATION)
			mu = 1 - alpha
		end
		if (not sVal) or (alpha <= 0 and not bar.realValue) or (alpha >= 1 and bar.realValue) then
			if (bar:GetParent()[field] == bar) then -- make sure we didn't get reassigned somehow
				bar:SetAlpha(1)
				bar:Hide()
				bar.orientation = nil
				bar.stop = nil
				bar.smoothstop = nil
				if bar.realValue then
					if field == "healthFade" then
						bar:GetParent().healthBar:SetValue(bar.realValue)
					elseif field == "manaFade" then
						bar:GetParent().manaBar:SetValue(bar.realValue)
					end
				end
				bar.realValue = nil
				bar:GetParent()[field] = _delFrame(bar)
			end
			_frames[bar] = nil
		else
			cVal = bar:GetValue()
			cVal = CosineInterpolate(cVal,sVal,mu)
			bar:SetValue(cVal)
		end
	end
end

function PitBull_BarFader:OnInitialize()
	self.db = PitBull:AcquireDBNamespace("BarFader")
	PitBull:RegisterDefaults("BarFader", "profile", {
		color = {0, 0.8, 0, 0.5},
		groups = {
			['*'] = {},
		}
	})
end

function PitBull_BarFader:OnEnable()
	self:Hook(_PitBull_Mana, "UpdateMana")
	self:Hook(_PitBull_Health, "UpdateHealth")
	self:ScheduleRepeatingEvent("PitBull_BarFader",self.UpdateFaders,0,self)
end

function PitBull_BarFader:PositionHealthBar(frame)
	frame.healthFade:ClearAllPoints()
	if frame.healthFade.orientation == "HORIZONTAL" then
		frame.healthFade:SetPoint("TOPLEFT", frame.healthBar.bg, "TOPLEFT", 0, 0)
	else
		frame.healthFade:SetPoint("BOTTOMLEFT", frame.healthBar.bg, "BOTTOMLEFT", 0, 0)
	end
end

function PitBull_BarFader:OnClearUnitFrame(unit, frame)
	if frame.healthFade then
		_frames[frame.healthFade] = nil
		frame.healthFade = _delFrame(frame.healthFade)
	end
	if frame.manaFade then
		_frames[frame.manaFade] = nil
		frame.manaFade = _delFrame(frame.manaFade)
	end
end

function PitBull_BarFader:PositionManaBar(frame)
	frame.manaFade:ClearAllPoints()
	if frame.manaFade.orientation == "HORIZONTAL" then
		frame.manaFade:SetPoint("TOPLEFT", frame.manaBar.bg, "TOPLEFT", 0, 0)
	else
		frame.manaFade:SetPoint("BOTTOMLEFT", frame.manaBar.bg, "BOTTOMLEFT", 0, 0)
	end
end

function PitBull_BarFader:UpdateHealth(object, unit, frame)
	local healthvalue = frame.healthBar:GetValue()
	self.hooks[object].UpdateHealth(object,unit,frame)
	if self.db.profile.groups[frame.group].ignore then
		return
	end
	local db = _PitBull_Health.db.profile[frame.group]
	if (db.deficit == db.reverse) then
		if (frame.healthBar:GetValue() < healthvalue) then
			local healthFade
			if (frame.healthFade) then
				healthFade = frame.healthFade
			else
				healthFade = _newFrame("StatusBar", frame)
			end
			frame.healthFade = healthFade
			healthFade.realValue = nil
			healthFade:SetWidth(frame.healthBar:GetWidth())
			healthFade:SetHeight(frame.healthBar:GetHeight())
			healthFade.orientation = frame.healthBar:GetOrientation()
			healthFade:SetOrientation(healthFade.orientation)
			
			frame.healthBar:SetFrameLevel(3)
			healthFade:SetFrameLevel(2)
			
			local min, max = frame.healthBar:GetMinMaxValues()
			healthFade:SetMinMaxValues(min, max)
			healthFade:SetValue(healthvalue)
			healthFade:SetStatusBarTexture(PitBull:GetStatusBarTexture())
			local r, g, b = frame.healthBar:GetStatusBarColor()
			healthFade:SetStatusBarColor(r + 0.2, g + 0.2, b + 0.2)
			healthFade.faderstop = GetTime() + DURATION
			if (self.db.profile.groups[frame.group].smooth) then
				healthFade.smoothstop = frame.healthBar:GetValue()
				_framesSmooth[healthFade] = 'healthFade'
			else
				_frames[healthFade] = 'healthFade'
			end
			healthFade:Show()
			self:PositionHealthBar(frame)
		end
	else
		if frame.healthFade and frame.healthFade.realValue then
			healthvalue = frame.healthFade.realValue
		end
		if (frame.healthBar:GetValue() > healthvalue) then
			local healthFade
			if (frame.healthFade) then
				healthFade = frame.healthFade
			else
				healthFade = _newFrame("StatusBar", frame)
			end
			frame.healthFade = healthFade
			healthFade:SetWidth(frame.healthBar:GetWidth())
			healthFade:SetHeight(frame.healthBar:GetHeight())
			healthFade.orientation = frame.healthBar:GetOrientation()
			healthFade:SetOrientation(healthFade.orientation)
			frame.healthBar:SetFrameLevel(2)
			healthFade:SetFrameLevel(3)
			local min, max = frame.healthBar:GetMinMaxValues()
			healthFade:SetMinMaxValues(min, max)
			healthFade:SetValue(frame.healthBar:GetValue())
			
			-- Bad bad hacky stuff.  Let's not screw this up and break the real statusbars, k?
			healthFade.realValue = frame.healthBar:GetValue()
			frame.healthBar:SetValue(healthvalue)
			
			healthFade:SetAlpha(0)
			healthFade:SetStatusBarTexture(PitBull:GetStatusBarTexture())
			local r, g, b = frame.healthBar:GetStatusBarColor()
			healthFade:SetStatusBarColor(r, g, b)
			healthFade.faderstop = GetTime() + DURATION
			if (self.db.profile.groups[frame.group].smooth) then
				healthFade.smoothstop = healthFade.realValue
				_framesSmooth[healthFade] = 'healthFade'
			else
				_frames[healthFade] = 'healthFade'
			end
			healthFade:Show()
			self:PositionHealthBar(frame)
		end
	end
end

function PitBull_BarFader:UpdateMana(object, unit, frame)
	local manavalue = frame.manaBar:GetValue()
	self.hooks[object].UpdateMana(object,unit,frame)
	if self.db.profile.groups[frame.group].ignore then
		return
	end
	local db = _PitBull_Mana.db.profile[frame.group]
	if (db.deficit == db.reverse) then
		if (frame.manaBar:GetValue() < manavalue) then
			local manaFade
			if (frame.manaFade) then
				manaFade = frame.manaFade
			else
				manaFade = _newFrame("StatusBar", frame)
			end
			frame.manaFade = manaFade
			manaFade.realValue = nil
			manaFade:SetWidth(frame.manaBar:GetWidth())
			manaFade:SetHeight(frame.manaBar:GetHeight())
			manaFade.orientation = frame.manaBar:GetOrientation()
			manaFade:SetOrientation(manaFade.orientation)
			
			frame.manaBar:SetFrameLevel(3)
			manaFade:SetFrameLevel(2)
			
			local min, max = frame.manaBar:GetMinMaxValues()
			manaFade:SetMinMaxValues(min, max)
			manaFade:SetValue(manavalue)
			manaFade:SetStatusBarTexture(PitBull:GetStatusBarTexture())
			local r, g, b = frame.manaBar:GetStatusBarColor()
			manaFade:SetStatusBarColor(r + 0.2, g + 0.2, b + 0.2)
			manaFade.faderstop = GetTime() + DURATION
			if (self.db.profile.groups[frame.group].smooth) then
				manaFade.smoothstop = frame.manaBar:GetValue()
				_framesSmooth[manaFade] = 'manaFade'
			else
				_frames[manaFade] = 'manaFade'
			end
			manaFade:Show()
			self:PositionManaBar(frame)
		end
	else
		if frame.manaFade and frame.manaFade.realValue then
			manavalue = frame.manaFade.realValue
		end
		if (frame.manaBar:GetValue() > manavalue) then
			local manaFade
			if (frame.manaFade) then
				manaFade = frame.manaFade
			else
				manaFade = _newFrame("StatusBar", frame)
			end
			frame.manaFade = manaFade
			manaFade:SetWidth(frame.manaBar:GetWidth())
			manaFade:SetHeight(frame.manaBar:GetHeight())
			manaFade.orientation = frame.manaBar:GetOrientation()
			manaFade:SetOrientation(manaFade.orientation)
			frame.manaBar:SetFrameLevel(2)
			manaFade:SetFrameLevel(3)
			local min, max = frame.manaBar:GetMinMaxValues()
			manaFade:SetMinMaxValues(min, max)
			manaFade:SetValue(frame.manaBar:GetValue())
			
			-- Bad bad hacky stuff.  Let's not screw this up and break the real statusbars, k?
			manaFade.realValue = frame.manaBar:GetValue()
			frame.manaBar:SetValue(manavalue)
			
			manaFade:SetAlpha(0)
			manaFade:SetStatusBarTexture(PitBull:GetStatusBarTexture())
			local r, g, b = frame.manaBar:GetStatusBarColor()
			manaFade:SetStatusBarColor(r, g, b)
			manaFade.faderstop = GetTime() + DURATION
			if (self.db.profile.groups[frame.group].smooth) then
				manaFade.smoothstop = manaFade.realValue
				_framesSmooth[manaFade] = 'manaFade'
			else
				_frames[manaFade] = 'manaFade'
			end
			manaFade:Show()
			self:PositionManaBar(frame)
		end
	end
end

-- Global Options
local function getColor()
	return unpack(PitBull_BarFader.db.profile.color)
end
local function setColor(r,g,b,a)
	PitBull_BarFader.db.profile.color = {r,g,b,a}
end

-- Unit/Group Specific Options
local function getSmooth(group)
	return PitBull_BarFader.db.profile.groups[group].smooth
end
local function setSmooth(group, value)
	PitBull_BarFader.db.profile.groups[group].smooth = value
end
local function getIgnored(group)
	return PitBull_BarFader.db.profile.groups[group].ignore
end
local function setIgnored(group, value)
	PitBull_BarFader.db.profile.groups[group].ignore = value
end
PitBull_BarFader:RegisterPitBullOptionsMethod(function(group)
	return {
		name = L["BarFader"],
		desc = L["BarFader"],
		type = 'group',
		args = {
			ignore = {
				type = 'toggle',
				name = L["Ignore Unit(s)"],
				desc = L["Ignore Unit(s)"],
				get = getIgnored,
				set = setIgnored,
				passValue = group,
			},
			smooth = {
				type = 'toggle',
				name = L["Smooth Value Changing"],
				desc = L["Smooth Value Changing"],
				get = getSmooth,
				set = setSmooth,
				passValue = group,
			},
		}
	}
end)

_wrap(PitBull_BarFader, 'BarFader')
